/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 2001 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 *
 * --------------------------------------------------------------------------*/

#include "jabberd.h"

#define N_ERROR -1
#define N_NONE 0
#define N_READY 1

#define MAX_NODES 256

typedef struct master_inst_st *master_inst, _master_inst;

typedef struct farm_node_st
{
    pool p;
    master_inst fi;
    char *id; /* stream id */
    mio m; /* our connection to the node */
    int state;
    char *host; /* ip addr to redirect to */
    int load; /* number of sessions, not pending sessions */
    int index; /* our place in node_tab */
    struct farm_node_st *next;
} *farm_node, _farm_node;

typedef struct
{
    pool p;
    char *user;
    farm_node node;
    int count;
} *session, _session;

struct master_inst_st
{
    instance i;
    xdbcache xc;
    mio m; /* accepts connections from nodes */
    char *secret;
    char *ip;
    int port;
    char *route_attrib;
    int broadcast;
    farm_node nodes;
    farm_node node_tab[MAX_NODES];
    int node_count;
    xht hosts; /* hostnames we condsider local to our farm */
    xht sessions; /* established sessions */
    xht pending; /* where we routed packets from a user without a session */
};

/* broadcast session start/end to other nodes */
void master_broadcast(master_inst fi, farm_node ignore, xmlnode x)
{
    if (fi->broadcast)
    {
        farm_node cur;

        xmlnode_hide_attrib(x,"from"); /* just incase */
        for (cur = fi->nodes; cur != NULL; cur = cur->next)
            if (cur != ignore)
                mio_write(cur->m,xmlnode_dup(x),NULL,-1);
    }
    xmlnode_free(x);
}

/* this probably isn't the best way to do this... but it works for the moment */
/* weighted nodes would be cool */
farm_node master_choose(master_inst fi)
{
    int i;
    i = rand() % fi->node_count;
    log_debug("xyz","Selected node %d, total %d",i,fi->node_count);
    return fi->node_tab[i];
}

void master_rm_node(master_inst fi, farm_node node)
{
    farm_node last;

    -- fi->node_count;
    last = fi->node_tab[fi->node_count];
    fi->node_tab[fi->node_count] = NULL; /* rm from end */

    /* take the last node in the table and replace removed node */
    fi->node_tab[node->index] = last;
    last->index = node->index;

    /* remove from table */
    if (fi->nodes != node)
    {
        farm_node cur;
        for (cur = fi->nodes; cur->next != node; cur = cur->next);
        cur->next = node->next;
    }
    else
        fi->nodes = node->next;
}

void master_add_node(master_inst fi, farm_node node)
{
    /* add to table */
    node->index = fi->node_count++;
    fi->node_tab[node->index] = node;

    /* add to list */
    node->next = fi->nodes;
    fi->nodes = node;
}

/* remove all sessions of a dead node */
void master_walking_dead(xht h, const char *key, void *val, void *arg)
{
    farm_node node = (farm_node) arg;
    session s = (session) val;

    if (s->node == node)
    {
        xmlnode x;

        x = xmlnode_new_tag("route");
        xmlnode_put_attrib(x,"to",s->user);
        xmlnode_put_attrib(x,"type","error");

        master_broadcast(node->fi,node,x);

        xhash_zap(h,key);
        pool_free(s->p);
    }
}

/* this is called when a node is removed, or its connection closed */
void master_node_dead(master_inst fi, farm_node node)
{
    master_rm_node(fi,node);
    /* yes, this isn't very efficient */
    xhash_walk(fi->sessions,master_walking_dead,(void *) node);
    xhash_walk(fi->pending,master_walking_dead,(void *) node);
    pool_free(node->p);
}

/* tries to look up a session for the corrosponding user */
session master_get_session(master_inst fi, jid id)
{
    session s;
    char *user;

    jid_set(id,NULL,JID_RESOURCE);
    user = jid_full(id);

    s = (session) xhash_get(fi->sessions,user);
    if (s == NULL)
    {
        log_debug(ZONE,"Checking pending sessions");
        s = (session) xhash_get(fi->pending,user);
    }

    return s;
}

session master_session_create(farm_node node, char *user)
{
    session s;
    pool p;

    p = pool_new();
    s = pmalloco(p,sizeof(_session));
    s->p = p;
    s->user = pstrdup(p,user);
    s->node = node;

    return s;
}

void master_route_session(farm_node node, xmlnode x)
{
    master_inst fi = node->fi;
    session s, pend;
    char *user = xmlnode_get_attrib(x,"to");

    pend = (session) xhash_get(fi->pending,user);
    if (pend != NULL)
    {

        if (--pend->count == 0)
        {
            xhash_zap(fi->pending,user);
            pool_free(pend->p);
        }
    }

    ++node->load;

    s = (session) xhash_get(fi->sessions,user);
    if (s == NULL)
    {
        log_debug(ZONE,"SESSION[%s] CREATED",user);

        s = master_session_create(node,user);
        s->count = 1;
        xhash_put(fi->sessions,s->user,(void *) s);
        master_broadcast(fi,node,x);
    }
    else
    {
        log_debug(ZONE,"SESSION[%s] INCREMENT %d",user,s->count);
        ++s->count;
        xmlnode_free(x);
    }
}

/* a session ended */
void master_route_error(farm_node node, xmlnode x)
{
    master_inst fi = node->fi;
    session s;
    char *user = xmlnode_get_attrib(x,"to");

    s = (session) xhash_get(fi->sessions,user);
    if (s != NULL)
    {
        --node->load;
        if (--s->count == 0)
        {
            log_debug(ZONE,"SESSION[%s] END",user);
            xhash_zap(fi->sessions,user);
            pool_free(s->p);
            master_broadcast(fi,node,x);
            return;
        }
        else
        {
            log_debug(ZONE,"SESSION[%s] DECREMENT %d",user,s->count);
        }
    }
    else
    {
        log_debug(ZONE,"SESSION[%s] non-existent?",user);
    }
    xmlnode_free(x);
}

void master_route_farm(farm_node node, xmlnode x)
{
    node->host = pstrdup(node->p,xmlnode_get_attrib(x,"ip"));
    xmlnode_free(x);
    node->state = N_READY;
    master_add_node(node->fi,node);
}

void master_node_deliver(master_inst fi, jid id, xmlnode x)
{
    session s;
    farm_node node;

    if (fi->node_count == 0)
    {
        log_debug(ZONE,"No available nodes");
        jutil_error(x,(terror){500,"No available nodes"});
        deliver(dpacket_new(x),fi->i);
        return;
    }

    /* packet comming from s2s(or other services) are routed to nodes base on the to="" attribute */
    if (id->user == NULL || (s = master_get_session(fi,id)) == NULL)
    {  /* the user doesn't have a session or this packet is directed to the server. */
        node = master_choose(fi);

        if (id->user)
        {
            s = master_session_create(node,jid_full(id));
            xhash_put(fi->pending,s->user,(void *) s);
            ++s->count;
        }

        /* the user is offline, or this packet is directed to the server. */
        log_debug(ZONE,"Delivering offline packet");
    }
    else
    {
        node = s->node;
        log_debug(ZONE,"Delivering to node %X",node);
    }

    /* send on... */
    mio_write(node->m,x,NULL,-1);
}

void master_normal(master_inst fi, xmlnode x)
{
    pool p;
    jid id;

    p = xmlnode_pool(x);
    /* id = jid_new(p,xmlnode_get_attrib(x,fi->route_attrib)); */
    id = jid_new(p,xmlnode_get_attrib(x,"to"));

    if (xhash_get(fi->hosts,id->server))
        master_node_deliver(fi,id,x); /* it's to another node */
    else
        deliver(dpacket_new(x),fi->i);
}

/* packets from nodes are handled here */
void master_node_packets(farm_node node, xmlnode x)
{
    if ((*xmlnode_get_name(x)) != 'r' || xmlnode_get_attrib(x,"from"))
    {
        master_normal(node->fi,x);
    }
    else
    {
        char *type = xmlnode_get_attrib(x,"type");

        if (type == NULL)
        {
            log_debug(ZONE,"Unknown route type?");
            xmlnode_free(x);
            return;
        }

        switch (*type)
        {
        case 's': /* session */
            master_route_session(node,x);
            break;
        case 'e': /* error */
            master_route_error(node,x);
            break;
        case 'f': /* farm */
            master_route_farm(node,x);
            break;
        default:
            log_debug(ZONE,"Unknown route type?");
            xmlnode_free(x);
            break;
        }
    }
}

void master_handshake(farm_node node, xmlnode x)
{
    master_inst fi = node->fi;
    mio m = node->m;
    pool p;
    char hashbuf[41];

    if(j_strcmp(xmlnode_get_name(x),"handshake") != 0)
    {
        mio_write(m,NULL,"<stream:error>Must send handshake first.</stream:error>",-1);
        mio_close(m);
        xmlnode_free(x);
        return;
    }

    p = xmlnode_pool(x);
    shahash_r(spools(p,node->id,fi->secret,p),hashbuf);
    if(j_strcmp(hashbuf,xmlnode_get_data(x)) == 0)
    {
        log_debug(ZONE,"Negoiated new node %s",node->id);

        node->state = N_READY;
        master_add_node(fi,node);
        mio_write(m,NULL,"<handshake/>",-1);
    }
    else
    {
        mio_write(m,NULL,"<stream:error>Invalid handshake</stream:error>",-1);
        mio_close(m);
    }
    xmlnode_free(x);
}

/* Handle incoming packets from the xstream associated with an MIO object */
void master_node_read(mio m, int state, void* arg, xmlnode x)
{
    farm_node node = (farm_node) arg;
    master_inst fi = node->fi;

    switch(state)
    {
    case MIO_XML_ROOT:
        if (j_strcmp(xmlnode_get_attrib(x,"xmlns"),"jabber:component:farm") != 0)
        {
            log_debug(ZONE,"Received invalid namespace");
            mio_write(m,NULL,SERROR_NAMESPACE,-1);
            mio_close(m);
        }
        else
        {
            xmlnode hd;

            hd = xstream_header("jabber:component:farm",NULL,fi->i->id);
            node->id = pstrdup(node->p,xmlnode_get_attrib(hd,"id"));
            mio_write(m,NULL,xstream_header_char(hd),-1);
            xmlnode_free(hd);
        }
        xmlnode_free(x);
        break;

    case MIO_XML_NODE:
        if(node->state == N_READY)
            master_node_packets(node,x);
        else
            master_handshake(node,x);
        break;

    case MIO_ERROR:
        log_debug(ZONE,"node socket error");
        node->state = N_ERROR;
        break;

    case MIO_CLOSED:
        log_debug(ZONE,"closing node socket");
        /* remove this node from the farm */
        master_node_dead(fi,node);
        break;

    default:
        break;
    }
}

void master_node_listen(mio m, int state, void *arg, xmlnode x);

void master_listen_again(void *arg)
{
    master_inst fi = (master_inst) arg;
    pth_nap(pth_time(2,0));
    if(mio_listen(fi->port,fi->ip,master_node_listen,(void *) fi,NULL,
                  mio_handlers_new(NULL,NULL,MIO_XML_PARSER)) == NULL)
    {
        log_error("farm","Unable to listen on the configured ip and port");
    }
}

void master_node_listen(mio m, int state, void *arg, xmlnode x)
{
    farm_node node;
    pool p;

    switch(state)
    {
    case MIO_NEW:
        /* we accepted a new node connection */
        p = pool_new();
        node = pmalloco(p,sizeof(_farm_node));
        node->fi = (master_inst) arg;
        node->p = p;
        node->m = m;
        node->state = N_NONE;

        mio_reset(m,master_node_read,(void *) node);
        break;

    case MIO_CLOSED:
        log_debug(ZONE,"listening socket closed");
        mtq_send(NULL,mio_pool(m),master_listen_again,arg);
        break;

    default:
        break;
    }
}

result master_phandler_normal(instance i, dpacket d, void *arg)
{
    master_inst fi = (master_inst) arg;

    switch (d->type)
    {
    case p_ROUTE:
    case p_NORM:
        if (xhash_get(fi->hosts,d->id->server) == NULL)
        {
            char *host = pstrdup(i->p,d->id->server);
            xhash_put(fi->hosts,host,(void *) host);
        }

        master_node_deliver(fi,jid_new(d->p,xmlnode_get_attrib(d->x,fi->route_attrib)),d->x);
        break;

    default:
        return r_ERR;
    }

    return r_DONE;
}

result master_debug(void *arg)
{
    master_inst fi = (master_inst) arg;
    farm_node cur;
    char buffer[1024];
    int pos = 0;

    memset(buffer,0,sizeof(buffer));

    for (cur = fi->nodes; cur != NULL; cur = cur->next)
    {
        sprintf(buffer + pos,"Node %X: load %d\n",(unsigned int) cur,cur->load);
        pos += strlen(buffer + pos);
    }
    log_debug(fi->i->id,"farm debug: %s",buffer);
    return r_DONE;
}

result master_timeout(void *arg)
{
    /* XXX implement pending session timeout */
    return r_DONE;
}

void master_init(instance i, xmlnode x)
{
    master_inst fi;
    xmlnode cfg;
    int error = 0;

    log_debug(ZONE,"farm initializing for section '%s'",i->id);

    fi = pmalloco(i->p,sizeof(_master_inst));
    fi->i = i;
    fi->xc = xdb_cache(i);

    cfg = xdb_get(fi->xc,jid_new(xmlnode_pool(x),"config@-internal"),"jabber:config:farm");
    if (cfg == NULL)
    {
        log_error("farm","Configuration not found!");
        return;
    }

    fi->secret = xmlnode_get_tag_data(cfg,"secret");
    fi->port = j_atoi(xmlnode_get_tag_data(cfg,"port"),0);
    fi->ip = xmlnode_get_tag_data(cfg,"ip");

    if (fi->secret == NULL)
    {
        log_error("farm","No secret configured");
        error = 1;
    }

    if (fi->port == 0)
    {
        log_error("farm","No port configured");
        error = 1;
    }

    if (error) return;

    if(!mio_listen(fi->port,fi->ip,master_node_listen,(void *) fi,NULL,mio_handlers_new(NULL,NULL,MIO_XML_PARSER)))
    {
        log_error("farm","Unable to listen on the configured ip and port");
        return;
    }

    fi->pending = xhash_new(128);
    fi->sessions = xhash_new(1024);
    fi->hosts = xhash_new(5);
    fi->broadcast = (xmlnode_get_tag(cfg,"broadcast") != NULL);

    if (xmlnode_get_tag_data(cfg,"route"))
        fi->route_attrib = pstrdup(i->p,xmlnode_get_tag_data(cfg,"route"));
    else
        fi->route_attrib = pstrdup(i->p,"to");

    srand(time(NULL));

    register_phandler(i,o_DELIVER,master_phandler_normal,(void *) fi);
    register_beat(j_atoi(xmlnode_get_tag_data(cfg,"delay"),60),master_timeout,(void *) fi);
    register_beat(15,master_debug,(void *) fi);
}
