/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 2001 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 *
 * --------------------------------------------------------------------------*/

#include "node.h"

#define c_CLOSED 0
#define c_OPEN 1
#define c_AUTHD 2

/* a session start on another node */
void node_rmts_start(fni ni, char *user)
{
    pool p;
    rmts r;
    modcall m;

    if (xhash_get(ni->remote_sessions,user) != NULL) abort();

    log_debug(ZONE,"REMOTE SESSION statred %s",user);

    p = pool_new();
    r = pmalloc(p,sizeof(_rmts));
    r->p = p;
    r->id = jid_new(p,user);
    r->ni = ni;

    xhash_put(ni->remote_sessions,jid_full(r->id),(void *) r);

    /* call remote_session_start cb (if any).  node_transport can ignore, remote sessions
       useless because inter-node routeing isn't required */

    m = ni->events[SESSION_START];
    if (m) (*(node_session_cb)m->cb) (r,m->arg);
}

void node_rmts_end(fni ni, char *user)
{
    /* a session ended on another node */
    rmts r;

    r = (rmts) xhash_get(ni->remote_sessions,user);
    if (r != NULL)
    {
        modcall m;

        log_debug(ZONE,"REMOTE SESSION ended %s",user);

        m = ni->events[SESSION_END];
        if (m) (*(node_session_cb)m->cb) (r,m->arg);
        xhash_zap(ni->remote_sessions,user);
        pool_free(r->p);
    }
    else
    {
        log_debug(ZONE,"Session end for remote session we know nothing about?");
    }
}

result node_call_packet(fni ni, int event, dpacket d)
{
    modcall m;

    m = ni->events[event];
    if (m == NULL)
        return r_PASS;

    return (*(node_dpacket_cb)m->cb) (ni,d,m->arg);
}

void node_packets(fni ni, xmlnode x)
{
    /* no from attribute, it's master to node packet not ment to be delivered */
    if (*(xmlnode_get_name(x)) != 'r' || xmlnode_get_attrib(x,"from") != NULL)
    {
        dpacket d = dpacket_new(x);

        log_debug(ZONE,"normal packet from master");
        if (node_call_packet(ni,PACKET_FARM,d) != r_DONE)
        {
            deliver(d,ni->i);
        }
        return;
    }

    switch (*xmlnode_get_attrib(x,"type"))
    {
    case 's': /* session */
        node_rmts_start(ni,xmlnode_get_attrib(x,"to"));
        break;
    case 'e': /* error */
        node_rmts_end(ni,xmlnode_get_attrib(x,"to"));
        break;
    default:
        log_debug(ZONE,"Unknown route type");
        break;
    }

    xmlnode_free(x);
}

void node_read(mio m, int state, void* arg, xmlnode x);

void *node_connect(void *arg)
{
    fni ni = (fni) arg;

    pth_nap(pth_time(2,0));
    mio_connect(ni->ip,ni->port,node_read,(void *) ni,ni->timeout,NULL,
                mio_handlers_new(NULL,NULL,MIO_XML_PARSER));
    return NULL;
}

void node_read(mio m, int state, void* arg, xmlnode x)
{
    fni ni = (fni) arg;
    xmlnode cur;
    char  hashbuf[41];

    log_debug(ZONE,"process XML: m:%X state:%d, arg:%X, x:%X",m,state,arg,x);

    switch (state)
    {
    case MIO_NEW:
        ni->state = c_OPEN;
        ni->m = m;

        /* Send a stream header to the server */
        cur = xstream_header("jabber:component:farm",ni->i->id,NULL);
        mio_write(m,NULL,xstream_header_char(cur),-1);
        xmlnode_free(cur);
        break;

   case MIO_XML_ROOT:
       /* Extract stream ID and generate a key to hash */
       shahash_r(spools(x->p,xmlnode_get_attrib(x,"id"),ni->secret,x->p),hashbuf);

       /* Build a handshake packet */
       cur = xmlnode_new_tag("handshake");
       xmlnode_insert_cdata(cur,hashbuf,-1);

       /* Transmit handshake */
       mio_write(m,NULL,xmlnode2str(cur),-1);

       xmlnode_free(cur);
       xmlnode_free(x);
       break;

    case MIO_XML_NODE:
        if (ni->state == c_AUTHD)
            node_packets(ni,x);
        else
        {
            if (j_strcmp(xmlnode_get_name(x), "handshake") == 0)
                ni->state = c_AUTHD;
            else
            {
                log_debug(ZONE,"Handshake failed!");
            }
            xmlnode_free(x);
        }
        break;

    case MIO_CLOSED:
        ni->state = c_CLOSED;
        log_debug(ZONE, "Attempting to reconnect...");
        pth_spawn(PTH_ATTR_DEFAULT,node_connect,(void *) ni);
        break;
    }
}

result node_phandler(instance i, dpacket d, void *arg)
{
    fni ni = (fni) arg;
    result r;

    r = node_call_packet(ni,PACKET_DELIVER,d);
    if (r == r_PASS)
    {
        node_route(ni,d->x);
        return r_DONE;
    }

    return r;
}

void node_register(fni ni, int event, void *cb, void *arg)
{
    modcall m;
    pool p;

    p = pool_new();
    m = pmalloco(p,sizeof(_modcall));
    m->p = p;
    m->cb = cb;
    m->arg = arg;

    ni->events[event] = m;
}

void node_session_start(fni ni, jid id)
{
    xmlnode x;

    /* notify other nodes and/or master */
    log_debug(ZONE,"SESSION:START [%s]",jid_full(id));

    x = xmlnode_new_tag("route");
    xmlnode_put_attrib(x,"to",jid_full(jid_user(id)));
    xmlnode_put_attrib(x,"type","session");

    mio_write(ni->m,x,NULL,-1);
}

void node_session_end(fni ni, jid id)
{
    xmlnode x;

    log_debug(ZONE,"SESSION:END [%s]",jid_full(id));

    x = xmlnode_new_tag("route");
    xmlnode_put_attrib(x,"to",jid_full(jid_user(id)));
    xmlnode_put_attrib(x,"type","error");

    mio_write(ni->m,x,NULL,-1);
}

/* in the future, nodes might maintain a connection with each other and
   send packet directly. this should speed up inter-node routing some */
void node_inter_route(fni ni, xmlnode x)
{
    mio_write(ni->m,x,NULL,-1);
}

/* send it to the master.  other services, etc.  A more optimized
   farm might only use node for inter-node routing, and base_accept/connect
   to a s2s process, with it's own meshed transports */
void node_route(fni ni, xmlnode x)
{
    mio_write(ni->m,x,NULL,-1);
}

xmlnode node_config(fni ni, char *tag)
{
    return xmlnode_get_tag(ni->cfg,tag);
}

typedef void (*modinit)(fni ni);

void node(instance i, xmlnode x)
{
    fni ni;
    xmlnode cfg, cur;
    modinit module;
    pool p = i->p;
    int error = 0;

    log_debug(ZONE,"node initializing for section '%s'",i->id);

    ni = pmalloco(p,sizeof(_fni));
    ni->i = i;
    ni->p = p;
    ni->xc = xdb_cache(i);

    cfg = xdb_get(ni->xc,jid_new(xmlnode_pool(x),"config@-internal"),"jabber:config:node");
    if (cfg == NULL)
    {
        log_error(i->id,"Configuration not found!");
        return;
    }


    ni->secret = pstrdup(p,xmlnode_get_tag_data(cfg,"secret"));
    ni->port = j_atoi(xmlnode_get_tag_data(cfg,"port"),0);
    ni->ip = pstrdup(p,xmlnode_get_tag_data(cfg,"ip"));

    if (ni->secret == NULL)
    {
        log_error(i->id,"No secret configured");
        error = 1;
    }
    if (ni->ip == NULL)
    {
        log_error(i->id,"No farm IP address configured");
        error = 1;
    }
    if (ni->port == 0)
    {
        log_error(i->id,"No port configured");
        error = 1;
    }

    if (error) return;

    ni->cfg = cfg;
    ni->timeout = j_atoi(xmlnode_get_tag_data(cfg,"timeout"),0);
    ni->remote_sessions = xhash_new(512);

    for(cur = xmlnode_get_firstattrib(x); cur != NULL; cur = xmlnode_get_nextsibling(cur))
    {
        if(j_strcmp(xmlnode_get_name(cur),"node") == 0)
            continue;

        /* vattrib is stored as firstchild on an attrib node */
        if((module = (modinit) xmlnode_get_firstchild(cur)) == NULL)
            continue;

        /* call this module for this session instance */
        log_debug(ZONE,"node: loading module %s",xmlnode_get_name(cur));
        (module)(ni);
        break; /* only load one module..? */
    }

    pth_spawn(PTH_ATTR_DEFAULT,node_connect,(void *) ni);
    register_phandler(i,o_DELIVER,node_phandler,(void *) ni);
}
